# File routines for input of shottab data from navigation/shot time
# text files.
#
# Since the style of input to these routines is extremely generic, these
# routines are built to be as flexible as possible with regards to the style
# of data input.
#
# We will allow input via the use of regular expressions:
#    this allows one to specify the format without regard to predefined ordering
#    of the data with great control over the format.
#
# Standard regular expressions will be defined here (and available via the GUI)
# such that the user need not worry about syntax.
#
# If more flexibility is required - they may use a custom regular expression
# (editable in the GUI) based on one of the existing ones.
#
# A good reference site for regular expressions in Python is:
#   http://py-howto.sourceforge.net/regex/regex.html
#
#
# Dalhousie University - Department of Oceanography.
#
# Author: Dave LeBlanc, dleblanc@cs.dal.ca, October 2002.


# This is the Python string interpolation module
# (for python 1.5 - but it works with 2.x)
from shottab.Itpl15 import Itpl

# regular expression module
import re


import os, mmap

# NOTE: we prefix these strings with 'r' such that they are raw - backslashes
# not considered escape sequences.
#
# NOTE: these are verbose regular expressions - a special (more) readable form
# of regular expressions.
#
#
def matchStr(strName):
    """Simply match the named string.  Matches >1 sequential number of word characters"""
    return r"(?P<" + strName + r">\w*)" + "\n"
    
def matchInt(intName):
    """
    Build a regular expression to match an integer with name intName
    will look like:
      (?P<intName>[+-]?\d+)
    This matches an optional sign in front of >= 1 digits.
    """
    return r"(?P<" + intName + r">[+-]?\d+)           # Read an integer value into " + \
           intName + "\n"

def matchFloat(floatName, strict=0):
    """
    Match a float with match name floatName. Matches an optional sign character,
    0 or more leading digits, a mandatory period (if strict is 1), and 0 or
    more optional trailing digits.
    
    NOTE: technically this means we could input ".", but this seems unlikely.

    Operation changed to default to non-strict mode (ie integer input cast to
    an int) so that heading can be input in either int or float format.
    """
    if strict:
        # strict - must have a decimal point
        return r"(?P<" + floatName + r">[+-]?\d*\.\d*)    # Read a float value into " + \
               floatName + "\n"
    else:
        # not strict (default) no decimal place required.
        return r"(?P<" + floatName + r">[+-]?\d*\.?\d*)    # Read a float value into " + \
               floatName + "\n"

def matchLatLon(coordName):
    """
    Match a lat/lon coordinate in either decimal degrees or decimal minute style.
    For this we use the regular expression OR (|) operator.  This matches either the
    first or second expressions within the (?:) field (unnamed group) - whichever one
    is given in the data, and returns this as the named expression match (coordName).
    
    For the first group (decimal minutes) we first match an integer - the degrees, and
    then match one or more characters of whitespace, followed by a float value (the
    decimal minutes).

    For the second group (decimal degrees) we simply match a float value - the
    decimal degree value.
    """
    matchStr = r"(?P<" + coordName + \
r""">                     # Read a lat/long value
  (?:[+-]?\d+\s+[+-]?\d*\.\d*\s*[NESWnesw]?) |  # match an integer, whitespace,
                                               # a float (decimal minutes), opt. whitespace
                                               # and an optional direction (NESW), _OR_
  (?:[+-]?\d*\.\d*)              # match just a leading float value (decimal degrees)
)
"""

    return matchStr


# A list of all regular expressions that we will use for shot time input - stored
# as a two-element tuple - the regular expression, and a comment to be displayed
# on the gui as to what it provides.  We use the helper functions above to prevent
# over duplication of text and improve it's ability to be edited by the user.
# NOTE: we are using Itpl to resolve $funcName() calls to actual function return
# values (eg matchInt() ).

shotTimeRegExpList = []

shotTimeRegExpList.append( \
("Standard odetics format - shotNo, day, hour, min, sec, partialSec, lat, lon",
r"""^\s*
$matchInt("shotNo")
\s+
$matchInt("day")
\s+
$matchInt("hour")
\s+
$matchInt("min")
\s+
$matchFloat("sec")
\s+
$matchStr("partialSec")
\s+
$matchLatLon("lat")
\s+
$matchLatLon("lon")"""))

shotTimeRegExpList.append( \
("shotNo, day, hour, min, sec, lat, lon - whitespace delimited",
r"""^\s*
$matchInt("shotNo")
\s+
$matchInt("day")
\s+
$matchInt("hour")
\s+
$matchInt("min")
\s+
$matchFloat("sec")
\s+
$matchLatLon("lat")
\s+
$matchLatLon("lon")
"""))

# Example regular expression with comments. The comments are vertically aligned in shotGui
shotTimeRegExpList.append( \
("Example with comments: fixed width, delimited",
r"""^\s*                                   #skip whitespace (if any) at start of line
(?P<shotNo>[\d\s]{4,4})   #reading variable 'shotNo' as min and max of 4 characters
\s{2,2}                               #skipping the next 2 whitespace characters
(?P<day>[\d\s]{3,3})         #reading variable 'day' as next 3 characters
\s*                                     #skipping any number of whitespace characters
(?P<hour>[\d\s]{2,2})        #reading variable 'hour' as next 2 characters
\s*
$matchInt("min")               #reading variable 'min' as next field (as integer of any length)
\s*
$matchFloat("sec")           #reading floating point number as variable 'sec'
\s*
$matchStr("partialSec")     #partialSec must be matched as a string.
\s*
$matchLatLon("lat")          #reading lat/long with special 'matchLatLong' syntax
\s*
$matchLatLon("lon")         #reading longitude
\s*
\d*                                     #example of skipping a number
"""))

# A similar list for input of navigation data
navRegExpList = []

navRegExpList.append( \
("day, hour, min, sec, lat, long, heading - in DDDHHMMSS LAT LON format",
r"""^
(?P<day>[\d\s]{3,3})
(?P<hour>[\d\s]{2,2})
(?P<min>[\d\s]{2,2})
(?P<sec>[\d\s]{2,2})
\s+
$matchLatLon("lat")
\s+
$matchLatLon("lon")
"""))

navRegExpList.append( \
("day, hour, min, sec, lat, long, heading - in DDDHHMMSS LAT LON HDG format",
r"""^
(?P<day>[\d\s]{3,3})
(?P<hour>[\d\s]{2,2})
(?P<min>[\d\s]{2,2})
(?P<sec>[\d\s]{2,2})
\s+
$matchLatLon("lat")
\s+
$matchLatLon("lon")
\s+
$matchFloat("heading")
"""))

navRegExpList.append( \
("day, hour, min, sec, lat, lon, heading - whitespace delimited",
r"""^\s*
$matchInt("day")
\s+
$matchInt("hour")
\s+
$matchInt("min")
\s+
$matchFloat("sec")
\s+
$matchLatLon("lat")
\s+
$matchLatLon("lon")
\s+
$matchFloat("heading")"""))


navRegExpList.append( \
("Standard odetics format - shotNo, day, hour, min, sec, partialSec, lat, lon",
r"""^\s*
\d*   # skipping the shot number
\s+
$matchInt("day")
\s+
$matchInt("hour")
\s+
$matchInt("min")
\s+
$matchFloat("sec")
\s+
$matchStr("partialSec")
\s+
$matchLatLon("lat")
\s+
$matchLatLon("lon")"""))


def strToLatLon(inStr):
    """
    combined with the matchLatLon() pattern, this routine converts a string representation
    of a lat/lon field in either decimal degree or decimal minute format (with optional
    NESW on the end) to a float value -- unifies all coordinate formats to decimal
    degree internally.
    """

    fields = inStr.split()
    if len(fields) >= 2:
        # If we reach here - coordinate is in degree minute format with an optional
        # direction (compass) character N/E/S/W.  We use the convention that
        # the southern and western hemispheres are negative, respective opposites
        # are positive.  This character must be used in place of a sign on the
        # coordinate values.
        #
        # It is a decimal second value - note that we handle the sign after the
        # fractional minute is added

        # A mapping of direction character to sign value
        dirMap = {'N' : 1.0, 'E': 1.0, 'S': -1.0, 'W': -1.0}

        signValue = 1.0
        if '-' in fields[0]:
            signValue = -1.0
            
        if len(fields) == 3:
            # If specified as a compass character, set neg truth value here
            signValue = dirMap[fields[2].upper()]

        # Else we simply keep the sign (signValue == 1.0)
            
        coordValue = (abs(float(fields[0])) + float(fields[1]) / 60.0) * signValue
    else:
        # Just a straight decimal degree value
        coordValue = float(inStr)
        
    return coordValue


def strToDate(inStr):
    """
    Convert a tuple represetation of a date, eg:
      (1, 2, 3, 4)
    into a tuple representing date as (day, hour, min, sec).

    The replace logic is far more readable than the equivalent
    regular expression.
    """
    clean = inStr.replace(",", " ")
    clean = clean.replace("(", "")
    clean = clean.replace(")", "")
    
    # TODO: use my more generic string split here
    fields = clean.split()
    return (int(fields[0]), int(fields[1]), int(fields[2]), float(fields[3]))


# An exception class we use to pass up errors
class LazyError(Exception):
    pass

# Attempted read beyond end of file.
class LazyEOF(LazyError):
    pass

class LazyLineReader:
    """
    Implements a sequence-like facility of loading records from a file
    in a lazy fashion - such that only the records requested are loaded.

    The previously read records are cached in a hash table - the
    record index is used to look up the record.  This cache is cleared
    when it passes a certain size so that we dont grow it too
    large with big files. -- This cache is important with the running
    sum algorithm as it fetches records for the same item multiple
    times.

    Note that mmap is used to allow for efficient file processing (and
    no-copy overhead) for large files.

    This then allows us to load files of virtually unlimited length, and
    with a constant seek time to any part of the file.

    For sequential (or close) iteration, only readline() is used to go to
    the next element - when accessing in a random fashion, seekApproxLine()
    is used to seek to approximately the target line.  This is not exact
    as line lengths can change by a small amount, so any caller performing
    non-sequential access should be aware of this (and take it into account)
    such as the NavReader with it's binary search for nav neighbors.
    """

    # TODO:    (should have largefile support on recent linux machines, need to test on
    # files > 2gigs on windows - esp NT 4.x).

    # Number of elements to scan forward to judge the average line length
    AVG_NUM = 20

    # If the intended read position is within this many of the current line number
    # then simply skip ahead, instead of using approximate seek
    ADJ_READ_LENGTH = 10

    # Maximum number of elements to cache before clearing it
    MAX_CACHE_SIZE = 2000

    def __init__(self, regExp, valueTypes, fileName, skipLines = 0):

        # Open the file
        self.fileName = fileName
        self.fileObj = open(fileName, 'r')

        try:
            # mmap this file as read only
            self.file = mmap.mmap(self.fileObj.fileno(), os.path.getsize(fileName), \
                access=mmap.ACCESS_READ)
        except OverflowError:
            print "File too big for mmap, falling back to normal file access mode"
            self.file = self.fileObj

        # skip skipLines lines of header here...
        for i in range(skipLines):
            self.file.readline()

        # save initial file position through tell()
        self.startFilePos = self.file.tell()

        # Get the average line length count
        self.avgLineLen = 0
        for i in range(self.AVG_NUM):
            self.avgLineLen += len(self.file.readline())
        self.avgLineLen = self.avgLineLen / float(self.AVG_NUM)

        # reset back to original position
        self.file.seek(self.startFilePos)

        # Expand (Itpl) and compile the regular expression
        # Beware - regExp is unicode, and needs to be casted to string.
        # and Itpl returns an object which also needs to be converted to string.
        self.regExp = re.compile(str(Itpl(str(regExp))), re.IGNORECASE | re.VERBOSE)

        self.valueTypes = valueTypes

        # A buffer of already converted lines (into records)
        self.cache = {}

        # The line number of the next to be read line
        self.curLineNo = 0

    def seekApproxLine(self, lineNo):
        """
        Seek (approximately) to the given line number.  This is
        approximate because the length of lines can change - thus we
        should be in the vicinity of the target line.  (this is
        ok given our application's needs (nav data neighbor
        searching).

        This routine is only called when we need to seek backwards
        before the buffered area, or if we are to go forwards
        more than the length of the buffer.

        This means that for sequential (or near) access, this
        routine should not be called, but it will be for the
        binary search used in the neighbor search.
        """
        # Care taken not to seek -1 if lineNo is 0.

        # TODO: should see if we are trying to seek past the EOF,
        # and raise the appropriate exception.

        if lineNo > 0:
            filePos = self.startFilePos + long(self.avgLineLen * lineNo) - 1
        else:
            filePos = self.startFilePos + long(self.avgLineLen * lineNo)
            
        self.file.seek(filePos)

        if lineNo > 0:
            self.file.readline()
        
        self.curLineNo = lineNo

    def __getitem__(self, idx):
        """
        This is where we will load the requested record.
        
        Since we may be dealing with _very_ large indexes, we must
        use long indexes, and xrange() such that these do not cause
        problems.
        """
        
        # Ensure we are always dealing in long arithmetic
        idx = long(idx)

        # If it's cached - simply return that
        if self.cache.has_key(idx):
            return self.cache[idx]

        # If we are reading past our adjacent readline length
        # or backwards, use the approximate line seeking algorithm
        if idx > self.curLineNo + self.ADJ_READ_LENGTH or \
               idx < self.curLineNo:
            self.seekApproxLine(idx)


        # Else dont use approximate seek - just use a seqeuence of
        # readlines (not approximate - this is exact)
        # Skip over lines till we start buffering
        while self.curLineNo <= idx:
            line = self.file.readline()
            if not line:
                raise LazyEOF("read past end of file")
            self.curLineNo += 1


        # Now try to convert it to our desired format using the regular
        # expression
        result = self.regExp.match(line)
        if result is not None:

            # Get a dictionary (hash) that maps the search term eg: "shotNo" to
            # the value, eg : "3".
            dict = result.groupdict()
            lineData = []

            # For all of our items, try to convert it
            for (itemName, itemCastFunc) in self.valueTypes:

                # If it was not found, insert None, so our order is kept
                if dict.has_key(itemName):
                    lineData.append(itemCastFunc(dict[itemName]))
                else:
                    lineData.append(None)
        else:
            lineData = None

        # If the cache is too big, flush it
        if len(self.cache) > self.MAX_CACHE_SIZE:
            self.cache = {}

        # Cache this item
        self.cache[idx] = lineData

        # Return the data
        return lineData

    def getRelativeCompletion(self):
        """Return a float value that represents how far we are through the file"""
        return self.file.tell() / float(os.path.getsize(self.fileName))

    def __len__(self):
        """
        Attempt to approximate the number of lines in the file.
        NOTE: this is not necessarily an accurate count of the
        number of lines in the file (could be off for large files) -
        too large or too small, thus one should iterate until
        receiving an appropriate LazyError to ensure all lines
        are read.
        """
        return int((os.path.getsize(self.fileName) - \
                     self.startFilePos) / self.avgLineLen)

