# Defines various components used in shotGui.py
#
# These include:
#
#    A Dialog to edit the format and name (and # lines to skip) of an input text file.
#
#    A text input validator for numeric input (actually generic string input) - both
#       integer and floating point.
#
#    float, int, lat/lon, and time (day,hour,min,sec) editor controls (built using
#       the above validator)
#
#
# Dalhousie University - Department of Oceanography.
#
# Author: Dave LeBlanc, dleblanc@cs.dal.ca, October 2002.

import os, string
from wxPython.wx import *

# The help for the regular expressions.
import shottab.help as help

# For automatic list control resizing
from wxPython.lib.mixins.listctrl import wxColumnSorterMixin, wxListCtrlAutoWidthMixin

# Import the string interpolation (expansion) routines
from shottab.Itpl15 import Itpl

import re
from shottab.fileio.fileio import *

class NumValidator(wxPyValidator):
    """
    Validate input so it conforms to a valid value.
    Uses a cast function - eg: int(), to attempt to convert the value, and catches ValueError
    exceptions.

    This is really something wxWindows should have more integrated into the API - like
    TrollTech's Qt (QDoubleValidator).
    """
    def __init__(self, castFunc, okChars, **args):
        # Call parent constructor with passed-on values
        wxPyValidator.__init__(self, *args)
        self.castFunc = castFunc
        self.okChars = okChars
        EVT_CHAR(self, self.OnChar)
    
    def Clone (self): 
        return NumValidator(self.castFunc, self.okChars)
       

    def isValid(self, str):
        try:
            # This is a special case we have to handle (leading +/-)
            if len(str) == 1 and str[0] in self.okChars:
                return 1
            
            value = self.castFunc(str)
            return 1
        except ValueError:
            return 0        

    def Validate(self, window):        # (window is parent window for control in question)
        ctrl = wxPyTypeCast(self.GetWindow(), "wxTextCtrl")
        return self.isValid(ctrl.GetValue())


    def OnChar(self, event):
        """Beep (and ignore) key if it is not in valid set of characters"""
        key = event.KeyCode()

        # don't ignore special characters
        if key < WXK_SPACE or key == WXK_DELETE or key > 255:
            event.Skip()
            return

        
        if chr(key) in self.okChars:
            # Ensure it is all now valid - we insert the char in the proper
            # position (in a temp string) and ensure this string is valid.

            ctrl = wxPyTypeCast(self.GetWindow(), "wxTextCtrl")
            pos = ctrl.GetInsertionPoint()
            
            # Since we want to get all of the text entered to now - not just any
            # valid portion, get the string edit value.
            testStr = ctrl.GetEditValue()
            
            # Add the new char in the proper position (with slices)
            testStr = testStr[:pos] + chr(key) + testStr[pos:]

            if self.isValid(testStr):
                event.Skip()
                return

        # If we reached here - the key is invalid - beep and exit.
        # Returning without calling Skip() eats the event before it
        # gets to the text control
        if not wxValidator_IsSilent():
            wxBell()

        return

    # Provide default (null) transfer methods
    def TransferToWindow(self):
        return 1
    def TransferFromWindow(self):
        return 1


# Below are a couple of helper classes that subclass wxTextCtrl with appropriate options
# for int/float editing using the above validators.

class IntEditCtrl(wxTextCtrl):
    def __init__(self, parent, id = -1, value = 0, pos = wxDefaultPosition, \
                 size = wxDefaultSize, style = 0, name = "IntEditCtrl"):
        
        # Call parent constructor with passed-on values
        wxTextCtrl.__init__(self, parent, id, str(value), pos, size, \
                               style, NumValidator(int, string.digits + "+-"), \
                               name)

    def SetValue(self, intVal):
        wxTextCtrl.SetValue(self, str(intVal))
        
    def GetEditValue(self):
        return wxTextCtrl.GetValue(self)
    
    def GetValue(self):
        # NOTE: if we only have a leading +/-, then we are not valid.
        # should return 0 in this case.
        intStr = self.GetEditValue()
        if len(intStr) == 1 and intStr[0] in "+-":
            return 0
        else:
            return int(intStr)


class FloatEditCtrl(wxTextCtrl):
    def __init__(self, parent, id = -1, value = 0.0, pos = wxDefaultPosition, \
                 size = wxDefaultSize, style = 0, name = "FloatEditCtrl"):
        
        # Call parent constructor with passed-on values
        wxTextCtrl.__init__(self, parent, id, str(value), pos, size, \
                               style, NumValidator(float, string.digits + "+-."), \
                               name)

    def SetValue(self, floatVal):
        wxTextCtrl.SetValue(self, str(floatVal))
        
    def GetEditValue(self):
        return wxTextCtrl.GetValue(self)

    def GetValue(self):
        """As above, if we only have a fragment of a valid number - return 0.0"""
        floatStr = self.GetEditValue()
        if len(floatStr) == 1 and floatStr[0] in "+-.":
            return 0.0
        else:
            return float(floatStr)



class LatLonCtrl(wxPanel):
    """
    A small component which allows one to edit Lat/Lon coordinates (with proper validation).
    Currently only supports decimal degree input.
    """
    def __init__(self, parent, value=(0.0, 0.0)):
        wxPanel.__init__(self, parent, -1)
        sizer = wxBoxSizer(wxHORIZONTAL)

        self.latEdit = FloatEditCtrl(self, value=str(value[0]))
        self.latEdit.SetSizeHints(-1, -1, maxW=60)
        
        self.lonEdit = FloatEditCtrl(self, value=str(value[1]))
        self.lonEdit.SetSizeHints(-1, -1, maxW=60)
        
        sizer.Add(wxStaticText(self, -1, "latitude:"), 0)
        sizer.Add(self.latEdit, 1)
        sizer.Add(wxStaticText(self, -1, "longitude:"), 0)
        sizer.Add(self.lonEdit, 1)                          

        self.SetSizer(sizer)
        sizer.Fit(self)

    def SetValue(self, coord):
        self.latEdit.SetValue(coord[0])
        self.lonEdit.SetValue(coord[1])
        
    def GetValue(self):
        """Return a tuple of (lat,lon) given the currently edited value"""
        return (float(self.latEdit.GetValue()), float(self.lonEdit.GetValue()))


class DateCtrl(wxPanel):
    """Edit a series of date values - in the following format:
        day: integer
        hour: integer
        minute: integer
        second: float
    """
    def __init__(self, parent, value=(0, 0, 0, 0.0)):
        wxPanel.__init__(self, parent, -1)

        sizer = wxBoxSizer(wxHORIZONTAL)

        self.dayEdit = IntEditCtrl(self, -1, value=str(value[0]))
        self.hourEdit = IntEditCtrl(self, -1, value=str(value[1]))
        self.minEdit = IntEditCtrl(self, -1, value=str(value[2]))
        self.secEdit = FloatEditCtrl(self, -1, value=str(value[3]))

        sizer.Add(wxStaticText(self, -1, "day:"), 0)
        sizer.Add(self.dayEdit, 1)
        sizer.Add(wxStaticText(self, -1, "hour:"), 0)
        sizer.Add(self.hourEdit, 1)
        sizer.Add(wxStaticText(self, -1, "minute:"), 0)
        sizer.Add(self.minEdit, 1)
        sizer.Add(wxStaticText(self, -1, "second:"), 0)
        sizer.Add(self.secEdit, 1)

        self.SetSizer(sizer)
        sizer.Fit(self)

    def SetValue(self, date):
        self.dayEdit.SetValue(date[0])
        self.hourEdit.SetValue(date[1])
        self.minEdit.SetValue(date[2])
        self.secEdit.SetValue(date[3])

    def GetValue(self):
        return (self.dayEdit.GetValue(), self.hourEdit.GetValue(), \
                self.minEdit.GetValue(), self.secEdit.GetValue())

    

class ListCtrlAutoWidth(wxListCtrl, wxListCtrlAutoWidthMixin):
    """
    Simple class to resize a list control automatically
    If this looks complicated, dont worry - we simply use it
    in place of a wxListCtrl below.
    """
    def __init__(self, parent, ID, pos=wxDefaultPosition,
                 size=wxDefaultSize, style=0):
        wxListCtrl.__init__(self, parent, ID, pos, size, style)
        wxListCtrlAutoWidthMixin.__init__(self)


class TextViewDialog(wxDialog):
    """
    A very simple dialog which shows some text in a scroll window, with
    an Ok button.
    Used for viewing the output from running the backend, viewing the
    expanded regular expression, etc.
    """
    OK_BUTTON_ID = wxNewId()
    def __init__(self, parent, title, textData, size=(500, 600)):

        # Call the parent constructor with appropriate values
        wxDialog.__init__(self, parent, -1, title, size=size, \
                          style=wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER | \
                          wxTE_DONTWRAP)
        self.textData = textData

        sizer = wxBoxSizer(wxVERTICAL)
        
        self.viewText = wxTextCtrl(self, -1, \
                                   style= wxTE_READONLY | wxTE_MULTILINE, \
                                   value=self.textData)
        self.viewText.SetBackgroundColour(wxLIGHT_GREY)
        
        # An ok button
        okButton = wxButton(self, self.OK_BUTTON_ID, "&Ok")
        EVT_BUTTON(self, self.OK_BUTTON_ID, self.okClicked)

        okButton.SetFocus()


        sizer.Add(self.viewText, 1, wxEXPAND)
        sizer.Add(okButton, 0, wxALIGN_CENTER)
        
        self.SetAutoLayout(1)        
        self.SetSizer(sizer)
        sizer.Fit(self)
        
        self.SetSize(size)
        self.Centre()

        self.Show(1) 

    def okClicked(self, evt):
        self.EndModal(wxID_OK)
        
    def addText(self, text):
        self.textData += text
        self.viewText.SetValue(self.textData)

class TextDataDialog(wxDialog):
    """
    Allows a user to describe columnar text data format to our program from virtually any
    variety of text file.  The data is read line by line using a regular expression
    - NOTE: we provide the following functions which get evaluated to form
    regular expression matches to the specified named group:

      matchInt(name), matchFloat(name), matchLatLon(name)

    This allows one to input "matchInt("foobar")" rather than the complicated regular
    expression that corresponds to this match (though they are still fully able to
    enter these kinds of regular expressions in this window).
    
    We also provide a preview pane - showing the values of the matched variables for each
    line.  This provides a quick way to ensure the regular expression is working correctly.
    
    Is a modal popup dialog allowing the user to edit the file name and style of data
    As with other dialogs, after the user presses OK or Cancel, the status and any
    related data will be available from this object (just as with a file dialog).

    If this class gets too big, it may be separated to a different file (module).

    This class is generic enough to be used in both the Shot Time and Navigation input
    file roles - so we customize to that role based upon some parameters passed to
    the constructor.
    """
    
    TEST_DATA_BUTTON_ID = wxNewId()
    REG_HELP_BUTTON_ID = wxNewId()
    REG_EXPANDED_BUTTON_ID = wxNewId()
    FILE_BUTTON_ID = wxNewId()    
    CANCEL_BUTTON_ID = wxNewId()
    OK_BUTTON_ID = wxNewId()
    REG_RADIO_ID = wxNewId()
    
    def __init__(self, parent, title, regExpList, valueTypes, initialData):

        # Call the parent constructor with appropriate values
        wxDialog.__init__(self, parent, -1, title, \
                          style=wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER )

        # Note that regExpList is a list of two-element tuples - the 1st element is
        # a comment describing the purpose of the regular expression, the second
        # is the regular expression itself.  This is the way these are stored in
        # fileio.py
        #
        # valueTypes is a list of 2-element tuples that holds the value name (string)
        # and a cast function - we only use the 1st name field.  (see fileio.py for
        # how this is used)
        
        self.regExpList = regExpList
        self.resultData = initialData
        self.valueTypes = valueTypes

        # Make a grid layout - labels on the left, edit controls on the right.
        # Items are placed in left to right, top down order (starting a new row after
        # every 2 items). -- Just as in main window.
        self.sizer = wxFlexGridSizer(cols = 2)

        # Designate row/col 1 (indexed from 0 - therefore the right column, 2nd row)
        # as growable.
        self.sizer.AddGrowableCol(1)
        self.sizer.AddGrowableRow(3)
        self.sizer.AddGrowableRow(4)

        self.sizer.Add(wxStaticText(self, -1, "Input File:"), 0, wxTOP, 5)

        fileSizer = wxBoxSizer(wxHORIZONTAL)
        # A label for the filename
        self.fileText = wxTextCtrl(self, -1, style=wxTE_READONLY, value=initialData[0])
        self.fileText.SetBackgroundColour(wxLIGHT_GREY)
        
        # A button to choose
        fileButton = wxButton(self, self.FILE_BUTTON_ID, "Choose...")
        EVT_BUTTON(self, self.FILE_BUTTON_ID, self.chooseFile)

        # Add both elements to the horizontal sizer
        fileSizer.Add(self.fileText, 1, wxEXPAND | wxTOP, 5)
        fileSizer.Add(fileButton, 0)

        self.sizer.Add(fileSizer, 1, wxEXPAND)

        # Number of lines to skip
        self.sizer.Add(wxStaticText(self, -1, "Lines of header to skip:"), 0)
        self.lineSkipSpin = wxSpinCtrl(self, -1, str(initialData[1]))
        self.sizer.Add(self.lineSkipSpin, 0)


        # Regular expression style selector
        self.sizer.Add(wxStaticText(self, -1, "Textual Format:"), 0)

        regDescriptions = []
        for (desc, unused) in regExpList:
            regDescriptions.append(desc)
            
        self.customRegExpIdx = len(regExpList)
        
        regDescriptions.append("Custom Regular Expression")
        
        self.textFormatRadio = wxRadioBox(self, self.REG_RADIO_ID, "", \
                                          choices = regDescriptions,
                                          majorDimension = 1, style = wxRA_SPECIFY_COLS)


        EVT_RADIOBOX(self, self.REG_RADIO_ID, self.regExpRadio)
        self.sizer.Add(self.textFormatRadio, 1, wxEXPAND)


        # Regular expression editor
        regEditBox = wxBoxSizer(wxVERTICAL)
        self.sizer.Add(wxStaticText(self, -1, "Regular Expression:"), 0)
        self.regExpEdit = wxTextCtrl(self, -1, style=wxTE_MULTILINE | wxTE_DONTWRAP)

        regEditBox.Add(wxStaticText(self, -1, \
                                    "This is a pre-expanded regular expression - see help"), 0)
        regHelpButton = wxButton(self, self.REG_HELP_BUTTON_ID, "Help")
        regOutputButton = wxButton(self, self.REG_EXPANDED_BUTTON_ID, "View Expanded")

        EVT_BUTTON(self, self.REG_HELP_BUTTON_ID, self.regExpHelp)
        EVT_BUTTON(self, self.REG_EXPANDED_BUTTON_ID, self.viewExpandedRegExp)
        
        regEditBox.Add(regHelpButton)
        regEditBox.Add(regOutputButton)        
        regEditBox.Add(self.regExpEdit, 1, wxEXPAND)
        self.sizer.Add(regEditBox, 1, wxEXPAND)        


        # assume they are using a custom regexp - select it for editing
        # - it will be enabled below in our call to regExpRadio()
        self.textFormatRadio.SetSelection(self.customRegExpIdx)
        if initialData[2] != "":
            self.regExpEdit.SetValue(initialData[2])
 
        # See if its not custom - and select this item.
        for i in range(len(regExpList)):
            if initialData[2] == regExpList[i][1]:
                # If this is in our regExp list - select it
                self.textFormatRadio.SetSelection(i)
            


        self.sizer.Add(wxStaticText(self, -1, "Input Data:"), 0)

        testDataSizer = wxBoxSizer(wxVERTICAL)
        testDataButton = wxButton(self, self.TEST_DATA_BUTTON_ID, "Test 1st 100 lines")
        EVT_BUTTON(self, self.TEST_DATA_BUTTON_ID, self.testData)
        
        self.dataList = ListCtrlAutoWidth(self, -1, style=wxLC_REPORT | \
                                   wxLC_HRULES | wxLC_VRULES)

        for i in range(len(self.valueTypes)):
            self.dataList.InsertColumn(i, self.valueTypes[i][0])            

        testDataSizer.Add(testDataButton)
        testDataSizer.Add(self.dataList, 1, wxEXPAND)

        self.sizer.Add(testDataSizer, 1, wxEXPAND)


        okCancelSizer = wxBoxSizer(wxHORIZONTAL)
        okButton = wxButton(self, self.OK_BUTTON_ID, "&Ok")
        EVT_BUTTON(self, self.OK_BUTTON_ID, self.okClicked)
        okButton.SetFocus()

        # Connect to the default cancel event handler
        cancelButton = wxButton(self, self.CANCEL_BUTTON_ID, "&Cancel")
        EVT_BUTTON(self, self.CANCEL_BUTTON_ID, self.cancelClicked)
        okCancelSizer.Add(okButton)
        okCancelSizer.Add(cancelButton)

        # Add a blank spacer
        self.sizer.Add(wxWindow(self, -1))
        
        self.sizer.Add(okCancelSizer, 0, wxALIGN_RIGHT)
        
        
        # Register the (grid) sizer
        self.SetAutoLayout(1)        
        self.SetSizer(self.sizer)
        self.sizer.Fit(self)

        self.SetSize((650, 600))
        self.Centre()
        self.Show(1)
        
        # simulate an event to the radio buttons so the state is in sync
        # and the regular expression edit box is populated
        self.regExpRadio(None)


    def testData(self, evt):
        """
        Populate the list with our input data (given the current regular
        expression) to test that elements are being input correctly.

        Only the 1st 100 elements are loaded
        """
        self.dataList.DeleteAllItems()                
        file = self.fileText.GetValue()
        skips = self.lineSkipSpin.GetValue()
        maxLines = 100
            
        try:
            # Try to open and parse the file - catch I/O errors (permissions, etc)
            # and regular expression parsing errors (below)
            
            data = LazyLineReader(self.regExpEdit.GetValue(), \
                                     self.valueTypes, file, skips)
            

        except IOError, errMsg:
            # File does not exist, etc
            dlg = wxMessageDialog(self, "Error: could not open %s:\n%s" % (file, str(errMsg)), \
                                  "File I/O Error", wxOK | wxICON_WARNING) 
            dlg.ShowModal() 
            dlg.Destroy() 
            return
            
        except re.error, errMsg:
            # Error in regular expression
            dlg = wxMessageDialog(self, "Error parsing regular expression:\n%s" % str(errMsg), \
                                  "Regular Expression Error", wxOK | wxICON_WARNING) 
            dlg.ShowModal() 
            dlg.Destroy() 
            return
            

        # Shoot some data onto the display - if the line fails to parse (blank line)
        # then post this.  Simply convert all items to string and show their value.
        try:
            
            for i in range(maxLines):
                line = data[i]
                
                if not line:
                    self.dataList.InsertStringItem(i, "Parse Error Line %d" % (i + 1))
                    continue
                
                self.dataList.InsertStringItem(i, str(line[0]))

                for j in range(1, len(line)):
                    self.dataList.SetStringItem(i, j, str(line[j]))
                    
        except LazyError:
            self.dataList.InsertStringItem(i, "File End/Input Error")
                

    def viewExpandedRegExp(self, evt):
        """View the resultant output from the expanded regular expression"""

        regExp = str(Itpl(str(self.regExpEdit.GetValue())))
        dialog = TextViewDialog(self, "Regular Expression Help", regExp)
        dialog.ShowModal()
        dialog.Destroy()

    def regExpHelp(self, evt):
        """Show a window containing help for regular expressions"""

        dialog = TextViewDialog(self, "Regular Expression Help", help.regExpHelpText)
        dialog.ShowModal()
        dialog.Destroy()

    def chooseFile(self, evt):
        """Choose an input file for the shot times"""
        (dir, file) = ("", "")
        if self.fileText.GetValue():
            (dir, file) = os.path.split(self.fileText.GetValue())
        
        fileDialog = wxFileDialog(self, "Choose File", style=wxOPEN, \
                                  defaultDir=dir, defaultFile=file)
        result = fileDialog.ShowModal()

        if result == wxID_OK:
            fileName = fileDialog.GetPath()

            # Now replace the control with the new pathname
            self.fileText.SetValue(fileName)

        fileDialog.Destroy()


    def cancelClicked(self, evt):
        self.EndModal(wxID_CANCEL)

    def okClicked(self, evt):
        # Fill out the results data field - ready to be grabbed by the caller
        
        if self.Validate():
            self.resultData = (self.fileText.GetValue(), self.lineSkipSpin.GetValue(), \
                               self.regExpEdit.GetValue())
            self.EndModal(wxID_OK)

    def getResult(self):
        # If this was not filled out correctly (by okClicked()) then it will simply return None.
        return self.resultData

    def regExpRadio(self, evt):
        """
        The user has selected a specific regular expression - we should
        clear the regExpEdit edit box an replace it with the regular
        expression to be used (if it is custom, we don't erase, as the user
        may wish to customize the expression).
        """
        selection = self.textFormatRadio.GetSelection()
        if selection == self.customRegExpIdx:
            self.regExpEdit.SetEditable(1)
            self.regExpEdit.SetBackgroundColour(wxWHITE)            
        else:
            self.regExpEdit.SetBackgroundColour(wxLIGHT_GREY)
            self.regExpEdit.SetEditable(0)

            # overwrite it with the new info
            self.regExpEdit.SetValue(self.regExpList[selection][1])
