# options.py - This defines the "Options" object - which is used to pass
# the values of specific options between the GUI editor and the batch
# shottab program.
#
# Loads and saves to a .ini style file.  Options are commented with
# any comments that are associated with them.
#
# Options class.
#
#   eg:
#    opts = Options("options.ini")
#    print(opts.exampleOption)
#
#  -- The above would load the options from the DEFAULT section of
#  options.ini and the print statement would print the value of the
#  "exampleOption" option.
#
#
# Dalhousie University - Department of Oceanography.
#
# Dave LeBlanc, dleblanc@cs.dal.ca, October 2002.

from ConfigParser import *
from types import *
import os, copy

class CaseConfigParser(ConfigParser):
    """
    Simply overrides optionxform() to return the option name in case sensitive form,
    such that we can use "shotNo" instead of having to lowercase everything.
    """
    def optionxform(self, optionstr):
        return optionstr

class InvalidValueError(Exception):
    pass
    
def customBool(value):
    boolMap = {"0" : False, "false" : False, 
        "1" : True, "true": True}
    lowValue = value.lower()
    if not (lowValue in boolMap):
        raise InvalidValueError("Error, invalid value for boolean: " + value)
    return boolMap[lowValue]
    
class Options:
    topComment = """\
# This is a configuration file for the shottab batch conversion program.  It is
# in a windows .ini style format. All options should be placed in the DEFAULT section.
# NOTE: If editing this by hand, please be sure to enter valid data for the types
# given - eg 0.0 for a float value, True/False for a boolean value, etc.
#
# NOTE: if editing a multi-line value, ensure the continuing lines are all prefixed
# by a leading tab (ensure it is a tab and not just spaces)
"""

    # This associates an option name with a comment to be inserted
    # into the configuration file.  If the option is not found here, no comment
    # is inserted.
    # This also specifies the order in which the options are output into the
    # file - top ones first.
    
    comments = (("cruiseID", "The name of the cruise - free form string"), \
                
                ("outputShotlogName", "The Filename to output the Shot Table information to"), \
                
                ("interpolatePositions", """Boolean value, False: use shot positions from the shotlog file
# True: interpolate positions from the navigation file"""), \
    
                ("interpolateHeadings", "Boolean value, if True: interpolate headings from nav data"), \
    
                ("shotTimeFile", "File name of the Shot Times input text file"), \
                               
                ("shotTimeSkipLines", "Integer, Number of lines of header of the shot time file to skip"), \
                                
                ("shotTimeRegExp", """Regular exression (in pre-expanded format) with which to
# filter the input file with  Please see the GUI for more information."""), \
                
                ("navFile", "The name of the navigation input file"), \
                          
                ("navSkipLines", "Number of lines of navigation file header to skip"), \
                               
                ("navRegExp", "The (pre-expanded) regular expression to load the nav file with"), \
                            
                ("correctForOffset", "Boolean, if true: correct shot positios for offset between GPS antenna and guns"), \

                ("makeRangePositive", "Boolean, if true: always make the range sign positive"), \
                                    
                ("obsClockResetTime", "(Day, hour, minute, second) time for obs clock reset"), \

                ("obsCalibrationTime", "Time of OBS calibration"), \

                ("obsDeploymentTime", "Time of OBS deployment"), \

                ("obsRecoveryTime", "Time of OBS recovery"), \

                ("obsLat", "OBS Latitude"), \

                ("obsLon", "OBS Longitude"), \

                ("centralMeridian", "Central Meridian"), \

                ("gunOffsetBehind", "Offset in meters of gun behind GPS antenna"), \

                ("gunOffsetRight", "Offset in meters of gun to the right of GPS antenna"), \

                ("clockDrift", "Total clock drift in msec"), \

                ("obsShipOffset", "Offset between ship and OBS clocks in msec"), \

                ("gunDelay", "Gun delay in milliseconds")
                )
    
    def __init__(self, cfgFileName = None):
        """Pulls all key/values from the given config file and sets
        the value of the named variable to that key/value.
        For now everything is in the [DEFAULT] section of this file.

        Below all defaults for needed attributes (options) are set,
        such that if they are not given in the file, they will still
        retain a default value.

        """

        # put all needed options here - their type is enforced upon load,
        # and the default here is used if not loaded.
        self.cruiseID = "Default Cruise ID"
        self.outputShotlogName = "shotlog.sht"
        self.interpolatePositions = False
        self.interpolateHeadings = False
        self.shotTimeFile = ""
        self.shotTimeSkipLines = 0
        self.shotTimeRegExp = ""
        self.navFile = ""
        self.navSkipLines = 0
        self.navRegExp = ""
        self.correctForOffset = False
        self.gunOffsetBehind = 0.0
        self.gunOffsetRight = 0.0
        self.makeRangePositive = False
        self.obsLat = 0.0
        self.obsLon = 0.0
        self.centralMeridian = 0
        
        self.obsClockResetTime = "(0, 0, 0, 0.0)"
        self.obsCalibrationTime = "(0, 0, 0, 0.0)"
        self.obsDeploymentTime = "(0, 0, 0, 0.0)"
        self.obsRecoveryTime = "(0, 0, 0, 0.0)"

        self.clockDrift = 0.0
        self.gunDelay = 0.0
        self.obsShipOffset = 0.0
        

        if cfgFileName != None:
            cfgFile = open(cfgFileName, "r+")

            cfg = CaseConfigParser()
            cfg.readfp(cfgFile)

            # For all the keys (option names) we create a new attribute and
            # set the value to the option's value.
            for (key, value) in cfg.defaults().items():
                if not hasattr(self, key):
                    raise InvalidValueError("Error: unknown configuration name: " + key)
            
                # A little test which ensures that we attempt to cast
                # the input (string) value to its appropriate type.
                try:
                    if type(getattr(self, key)) is BooleanType:
                        setattr(self, key, customBool(value))
                    elif type(getattr(self, key)) is IntType:
                        setattr(self, key, int(value))
                    elif type(getattr(self, key)) is LongType:
                        setattr(self, key, long(value))                
                    elif type(getattr(self, key)) is FloatType:
                        setattr(self, key, float(value))
                    else:
                        setattr(self, key, value)                

                except ValueError:
                    # The types that would fail - string would always pass.
                    typeNames = {IntType : "integer", LongType : "long integer",
                                 FloatType : "floating point"}
                    print("Error: invalid value for option \"%s\": \"%s\"" % (key, value))
                    print("Expected: %s value." % typeNames[type(getattr(self, key))])
                    raise InvalidValueError("Error: invalid value for option \"%s\": \"%s\"" \
                        % (key, value))
                
    def save(self, filename):
        """
        Save the options (attributes) to the given file name.  If a comment
        exists for this, it is printed above the option.

        NOTE: for line-continuation - have to replace \n with \n\t
        NOTE: we only deal with the [DEFAULT] section

        """
        newFile = open(filename, "w")
        newFile.write("%s\n\n" % self.topComment)
        newFile.write("[DEFAULT]\n")

        varDict = copy.copy(vars(self))

        # For all comments (in order)
        for (name, cmt) in self.comments:
            # If the attribute (option) exists - write it out
            if varDict.has_key(name):
                newFile.write("\n")
                newFile.write("# %s\n" % cmt)
                newFile.write("%s = %s\n" % (name, str(varDict[name]).replace('\n', '\n\t')))

                # Remove from our dictionary of attributes
                del varDict[name]


        # Now go through all the ones that werent commented (in undefined order)
        for (name, value) in varDict.items():
            newFile.write("\n")
            newFile.write("%s = %s\n" % (name, str(value).replace('\n', '\n\t')))


